/*****************************************************************************/
/*                                                                           */
/*                  Copyright (c) 1991-1994 by Jeffrey Sax                   */
/*                            All rights reserved                            */
/*                        Published and Distributed by                       */
/*                           Willmann-Bell, Inc.                             */
/*                             P.O. Box 35025                                */
/*                        Richmond, Virginia 23235                           */
/*                Voice (804) 320-7016 FAX (804) 272-5920                    */
/*                                                                           */
/*                                                                           */
/*                NOTICE TO COMMERCIAL SOFTWARE DEVELOPERS                   */
/*                                                                           */
/*        Prior to distributing software incorporating this code             */
/*        you MUST write Willmann-Bell, Inc. at the above address            */
/*        for validation of your book's (Astronomical Algorithms             */
/*        by Jean Meeus) and software Serial Numbers.  No additional         */
/*        fees will be required BUT you MUST have the following              */
/*        notice at the start of your program(s):                            */
/*                                                                           */
/*                    This program contains code                             */
/*              Copyright (c) 1991-1994 by Jeffrey Sax                       */
/*              and Distributed by Willmann-Bell, Inc.                       */
/*                         Serial #######                                    */
/*                                                                           */
/*****************************************************************************/

/* The following three lines tell the compiler to generate full */
/* function prototypes.  Erase or comment these lines if you    */
/* get an error message below. */

#ifndef PROTOTYPES
#define PROTOTYPES
#endif

/* First, some general constants and functions */

#ifdef _cplusplus
  #ifndef __cplusplus
    #define __cplusplus
  #endif
#endif

#ifdef __cplusplus

const double pi = 3.1415926535897932;
const double pi2 = (2 * pi);
const double DToR = (pi / 180);
const double HToR = (pi / 12);
const double RToD = (180 / pi);
const double RToH = (12 / pi);
const double SToR = (DToR / 3600);
const double EarthRadius = 6378.14;
const double SolarParallax = (8.794 * SToR);
const int FK4System = 1;
const int FK5System = 2;

const double secToT = (1 / (86400.0 * 36525.0));
const double JDJ2000 = 2451545;
const double JDB1950 = 2433282.4235;
const double TJ2000 = 0;
const double TB1950 = ((JDB1950 - JDJ2000) / 36525);

const double SiderealFactor = 1.0027379;
const double GaussConstant = 0.01720209895;
const double LightTimeConst = (0.0057755183 / 36525);
const double AU = 149597870.0;
const double Obl2000 = (23.43929111 * DToR);
const int EPJ2000 = 0;
const int EPDATE = 3;

#else

#define pi 3.1415926535897932
#define pi2 (2 * pi)
#define DToR (pi / 180)
#define HToR (pi / 12)
#define RToD (180 / pi)
#define RToH (12 / pi)
#define SToR (DToR / 3600)
#define EarthRadius 6378.14
#define SolarParallax (8.794 * SToR)
#define FK4System 1
#define FK5System 2

#define secToT (1 / (86400.0 * 36525.0))
#define JDJ2000 2451545
#define JDB1950 2433282.4235
#define TJ2000 0
#define TB1950 ((JDB1950 - JDJ2000) / 36525)

#define SiderealFactor 1.0027379
#define GaussConstant 0.01720209895
#define LightTimeConst (0.0057755183 / 36525)
#define AU 149597870.0
#define Obl2000 (23.43929111 * DToR)
#define EPJ2000 0
#define EPDATE 3

#endif

typedef double TVECTOR[3];   /*Rectangular coordinates*/
typedef struct tagTSVECTOR { /*Spherical coordinates*/
  double l, b, r;
  } TSVECTOR;

typedef double T2POLY[3];
typedef double T3POLY[4];
typedef double T4POLY[5];

typedef struct tagTDATE {
  int yy, mm;
  double dd;
  int h, m;
  double s;
  } TDATE;


/*************************\
* Basic support functions *
\*************************/

#ifdef __cplusplus
extern "C" {
#endif

#ifdef PROTOTYPES

double modpi2(double x);
double modpi(double x);
double ip(double x);

void SphToRect(TSVECTOR *S, TVECTOR P);
void RectToSph(TVECTOR P, TSVECTOR *S);
void XRotate(TVECTOR v, double c, double s, TVECTOR w);
void ZRotate(TVECTOR v, double c, double s, TVECTOR w);

void CalcSinCosTab(double X, int Degree, double *SinTab, double *CosTab);
double Interpol3(double y1, double y2, double y3, double n);
double InvInterpol3(double y1, double y2, double y3, double y);
double EvalPoly(const double *P, int n, double x);

#else

double modpi2(), modpi(), ip();
void SphToRect(), RectToSph(), XRotate(), ZRotate(), CalcSinCosTab();
double Interpol3(), EvalPoly();

#endif

#ifdef __cplusplus
}
#endif

/***********************\
* Date / Time functions *
\***********************/

#ifdef __cplusplus
extern "C" {
#endif

#ifdef PROTOTYPES

double ApproxDeltaT(double T);
double CalendarToJD(TDATE *Date);
int DayOfWeek(double JD);
int DayOfYear(TDATE *Date);
void DateOfEaster(int Year, TDATE *Date);
double EquationOfTime(double T, double RA, double Obl, double NutLon);
void JDToCalendar(double JD, TDATE *Date);
double JDToT(double JD);
double SiderealTime(double T);
double TToJD(double T);
double YearToJD0(int Year);

#else

double ApproxDeltaT(), CalendarToJD();
int DayOfWeek(), DayOfYear();
void DateOfEaster();
double EquationOfTime();
void JDToCalendar();
double JDToT(), SiderealTime(), TToJD(), YearToJD0();

#endif

#ifdef __cplusplus
}
#endif

/***********************\
* Position calculations *
\***********************/

typedef struct tagTORBITEL {
  double T0, M0;
  double a, q, n, e;
  double LonPeri, ArgPeri, LonNode, Incl;
  double MagParam1, MagParam2;
  } TORBITEL;

typedef struct tagTORBITCON {
  double A, B, C;
  double a, b, c;
  } TORBITCON;

#define DUMMY_SATELLITE 4 /* Number of fifth 'dummy' satellite of Jupiter */

#ifdef __cplusplus
extern "C" {
#endif

#ifdef PROTOTYPES

void ObserverCoord(double Phi, double Height, double *RhoCosPhi, double *RhoSinPhi);
void InitPlanetaryTerms(char *);
void FreePlanetaryTerms(int Planet);
void FreeAllPlanetaryTerms(void);
void ReadPlanetaryTerms(int Planet);
void SunPosLo(double T, double Obl, TSVECTOR *S);
void SunPosHi(double T, int RefFrame, TSVECTOR *S);
void SunRectCo(double T, int RefFrame, double Obl, TVECTOR R);
void PlPosHi(int Planet, double T, TSVECTOR *S);
void PlanetPosHi(int Planet, double T, TSVECTOR *S);
void PlutoPos(double T, TSVECTOR *S);
void HelioToGeo(TSVECTOR *SHelio, TSVECTOR *SEarth, TSVECTOR *SGeo);
void ReduceEclElements(double T, double T1, TORBITEL *OrbitEl);
void QuickReduceEclElements(TORBITEL *OrbitEl);
void FK4FK5ReduceEclElements(TORBITEL *OrbitEl);
double Kepler(double M, double e);
double Parabola(double dT, double q);
double NearParabola(double dT, double q, double e);
void PlanetOrbitEl(double T, int Planet, int Equinox, TORBITEL *OrbitEl);
void PlanetPosLo(double T, int Planet, int Equinox, double Obl, TVECTOR R);
void CalcOrbitCon(TORBITEL *OrbitEl, double Obl, TORBITCON *OrbitCon);
void PosRectCo(double T, TORBITEL *OrbitEl, TORBITCON *OrbitCon, TVECTOR R);
void RectHelioToGeo(TVECTOR RHelio, TVECTOR RSun, TSVECTOR *SGeo);
void CalcMoonAngles(double T, double *Om, double *D, double *M, double *Md, double *F, double *E);
void MoonPos(double T, TSVECTOR *S);
void MoonPosHi(double T, TSVECTOR *S);

/* Jovian satellites */
void Io(double, TSVECTOR *);
void Europa(double, TSVECTOR *);
void Ganymede(double, TSVECTOR *);
void Callisto(double, TSVECTOR *);
void JSatEclipticPosition(int n, double T, TVECTOR v);
void JSatViewFrom(int n, TVECTOR v, TSVECTOR *vsGeo, TVECTOR vDummy, int bScale, TVECTOR w);

/* Binary Stars */
typedef struct {
  double P, T, a, e;
  double Incl, LonNode, LonPeri;
  } TBINARYORBITEL;

void PositionOfBinary(double Year, TBINARYORBITEL *El, double *Separation, double *PosAngle);
double ApparentEccentricity(TBINARYORBITEL *El);

#else

void ObserverCoord(), InitPlanetaryTerms(), FreePlanetaryTerms();
void FreeAllPlanetaryTerms(), ReadPlanetaryTerms();
void SunPosHi(), SunRectCo(), PlPosLo(), PlanetPosHi(), PlutoPos();
void HelioToGeo(), ReduceEclElements();
void QuickReduceEclElements(), FK4FK5ReduceEclElements();
double Kepler(), Parabola(), NearParabola();
void PlanetOrbitEl(), PlanetPosLo(), CalcOrbitCon();
void PosRectCo(), RectHelioToGeo();
void CalcMoonAngles(), MoonPos();
void JSatEclipticPosition(), JSatViewFrom(), JSatRectify();
void PositionOfBinary();
double ApparentEccentricity();

#endif

#ifdef __cplusplus
}
#endif

/****************************\
* Coordinate transformations *
\****************************/

/* RiseSet flags */
#define NO_RISE 1   /* object doesn't rise             */
#define NO_SET 2    /* object doesn't set              */
#define DAY_MINUS 4 /* time of rise is of previous day */
#define DAY_PLUS 8  /* time of set is of next day      */
#define ALWAYS_ABOVE 16 /* object is always above horizon  */
#define ALWAYS_BELOW 32 /* object is always below horizon  */

/* RiseSet standard heights */

#ifdef __cplusplus

const double h0Sun = (-0.8333 * DToR);
const double h0Planet = (-0.5667 * DToR);

#else

#define h0Sun (-0.8333 * DToR)
#define h0Planet (-0.5667 * DToR)

#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef PROTOTYPES

void FK5ToFK4(double *RA, double *Decl, double dRA, double dDecl);
void FK4ToFK5(double *RA, double *Decl, double dRA, double dDecl);
void Aberration(double T, double Obl, int System, double *RA, double *Decl);
void EclAberration(double T, int System, double *l, double *b);
void eTermsOfAberration(int System, double *RA, double *Decl);
double AtmRefraction(double Altitude);
void EclToEqu(double l, double b, double Obl, double *RA, double *Decl);
void EquToHor(double RA, double Decl, double LST, double ObsLat, double *Az, double *Alt);
void EquToGal(double RA, double Decl, double *l, double *b);
double EquinoxCorrection(double T);
void NutationConst(double T, double *NutLon, double *NutObl);
void Nutation(double NutLon, double NutObl, double Obl, double *RA, double *Decl);
double Obliquity(double T);
void PrecessFK4(double T0, double T1, double *RA, double *Decl);
void PrecessFK5(double T0, double T1, double *RA, double *Decl);
double ParallacticAngle(double RA, double Decl, double ObsLat, double LAST);
void Parallax(double Parallax, double LST, double RhoCosPhi, double RhoSinPhi, double *RA, double *Decl);
void ParallaxHi(double sinParallax, double LST, double RhoCosPhi, double RhoSinPhi, double *RA, double *Decl);
void ParallaxHiAlt(double sinParallax, double LST, double RhoCosPhi, double RhoSinPhi, double *RA, double *Decl, double *SD);
double MoonSetHeight(double Parallax);
void RiseSet(double T, double DeltaT, double RA1, double Decl1, double RA2, double Decl2, double RA3, double Decl3, double Height0, double Lon, double Lat, double *Rise, double *Transit, double *Setting, int *Flags);

#else

void FK4ToFK5(), FK5ToFK4();
void Aberration(), EclAberration(), eTermsOfAberration();
double AtmRefraction();
void EclToEqu(), EquToHor();
double EquinoxCorrection();
void NutationConst(), Nutation();
double Obliquity();
void PrecessFK4(), PrecessFK5();
double ParallacticAngle();
void Parallax(), ParallaxHi(), ParallaxHiAlt();
double MoonSetHeight();
void RiseSet();

#endif

#ifdef __cplusplus
}
#endif

/**********************\
* Physical ephemerides *
\**********************/

typedef struct tagTMOONPHYSDATA {
  double l, b;      /* Total libration                       */
  double ld, bd;    /* Optical libration                     */
  double ldd, bdd;  /* Physical libration                    */
  double P;         /* Position angle                        */
  double i, k;      /* Phase angle and illuminated fraction  */
  double X;         /* Position angle of the bright limb     */
  } TMOONPHYSDATA;

typedef struct tagTSUNPHYSDATA {
  double P;         /* Position angle                      */
  double B0, L0;    /* Heliographic longitude and latitude */
  } TSUNPHYSDATA;

typedef struct tagTMARSPHYSDATA {
  double DS, DE;    /* Planetocentric declination of Sun and Earth */
  double Om;        /* Longitude of the central meridian           */
  double P;         /* Position angle of the axis                  */
  double Q, qq;     /* Position angle and -defect of illumination  */
  double d;         /* Angular diameter in seconds of arc          */
  } TMARSPHYSDATA;

typedef struct tagTJUPITERPHYSDATA {
  double DS, DE;    /* Planetocentric declination of Sun and Earth   */
  double Om1, Om2;  /* Longitude of the central meridians (I and II) */
  double C;         /* Correction for phase to the above             */
  double P;         /* Position angle of the axis                    */
  } TJUPITERPHYSDATA;

typedef struct tagTSATURNRINGDATA {
  double B, Bd;     /* Planetocentric declination of Earth and Sun */
  double P;         /* Position angle of the axis                  */
  double DeltaU;    /* Needed for magnitude calculation            */
  double aAxis, bAxis;
    /* Length of axes of rings in arcseconds       */
  } TSATURNRINGDATA;

#ifdef __cplusplus
extern "C" {
#endif

#ifdef PROTOTYPES

void MoonPhysEphemeris(double T, TSVECTOR *SGeo, TSVECTOR *SSun, 
  double Obl, double NutLon, double NutObl, 
  TMOONPHYSDATA *MoonPhysData);
double MoonSemiDiameter(double DistEarth);
void SunPhysEphemeris(double T, double l, double Obl, double NutLon, 
  TSUNPHYSDATA *SunPhysData);
double JDOfCarringtonRotation(int C);
int CarringtonRotation(double JD);
void MarsPhysEphemeris(double T, TSVECTOR *SHelio, TSVECTOR *SEarth, TSVECTOR *SGeo,
  double Obl, double NutLon, double NutObl,
  TMARSPHYSDATA *MarsPhysData);
void JupiterPhysEphemeris(double T, TSVECTOR *SHelio, TSVECTOR *SEarth, TSVECTOR *SGeo, 
  double Obl, double NutLon, double NutObl,
  TJUPITERPHYSDATA *JupiterPhysData);
void SaturnRing(double T, TSVECTOR *SHelio, TSVECTOR *SGeo, 
  double Obl, double NutLon, double NutObl,
  TSATURNRINGDATA *SaturnRingData);

/* Phase angles, magnitudes and semidiameters */
double CalcPhase(double rPS, double rES, double rPE);
double CalcPhaseAngle(double rPS, double rES, double rPE);
double PlanetMagnitude(int Planet, double rPS, double rPE, double i, double dU, double B);
double CometMagnitude(double rCS, double rCE, double g, double k);
double AsteroidMagnitude(double rAS, double rAE, double PhaseAngle, double H, double G);
double PlanetSemiDiameter(int Planet, double DistEarth, double *PolarDiam);
void CorrectSaturnSemiDiameter(double B, double *PolarSemiDiameter);

/* Stellar magnitudes */
double BrightnessToMag(double m);
double MagToBrightness(double b);
double AddMagnitudes(double m1, double m2);
double BrightnessRatio(double m1, double m2);

#else

void MoonPhysEphemeris(), SunPhysEphemeris();
double MoonSemiDiameter();
double JDOfCarringtonRotation();
int CarringtonRotation();
void MarsPhysEphemeris(), JupiterPhysEphemeris(), SaturnRing();

/* Phase angles, magnitudes and semidiameters */
double CalcPhase(), CalcPhaseAngle();
double PlanetMagnitude(), CometMagnitude(), AsteroidMagnitude();
double PlanetSemiDiameter();
void CorrectSaturnSemiDiameter();

/* Stellar magnitudes */
double Brightness(), Magnitude(), AddMagnitudes(), Brightnessratio();

#endif

#ifdef __cplusplus
}
#endif

/************************\
* Astronomical phenomena *
\************************/

/* EquinoxSolstice */

#define SPRINGEQUINOX 0
#define SUMMERSOLSTICE 1
#define AUTUMNEQUINOX 2
#define WINTERSOLSTICE 3

/* ConjunctionOpposition */

#define CONJUNCTION 1
#define OPPOSITION 0
#define INFCONJ 0
#define SUPCONJ 1
#define EASTELONGATION 0
#define WESTELONGATION 1

/* MoonPhase */

#define NEWMOON 0
#define FIRSTQUARTER 1
#define FULLMOON 2
#define LASTQUARTER 3

/* Eclipses */

#define TOTAL 1
#define ANNULAR 2
#define ANNULARTOTAL 3
#define TOTAL_NOT_CENTRAL 4
#define ANNULAR_NOT_CENTRAL 5
#define PARTIAL 6
#define OLDSYSTEM 1
#define PENUMBRAL 7
#define PARTPENUMBRAL 8

#ifdef __cplusplus
extern "C" {
#endif

#ifdef PROTOTYPES

double EquinoxSolstice(int Year, int Event);
int FirstkOfYear(int Year, int Planet, int Event);
double ConjunctionOpposition(int Planet, int Event, int k);
double GreatestElongation(int Planet, int Event, int k, double *Elongation);
double NextMoonPhase(double JD, int Event);
double NextSolarEclipse(double JD, double *Max, int *EclipseType);
double NextLunarEclipse(double JD, int System, double *Max, double *PartialSemiDuration, double *TotalSemiDuration, int *EclipseType);

#else

int FirstkOfYear();
double EquinoxSolstice(), ConjunctionOpposition(), GreatestElongation();
double MoonPhase(), NextSolarEclipse(), NextLunarEclipse();

#endif

#ifdef __cplusplus
}
#endif

/************************\
* Input/Output functions *
\************************/

/* The following 'magic' string clears */
/* the screen with or without an ANSI  */
/* screen driver installed.            */
#define CLSSTRING "\033[2J\014\b\040\b"

#ifdef __cplusplus
extern "C" {
#endif

#ifdef PROTOTYPES

/* Input functions */

int ReadInt(char *s);
double ReadReal(char *s);
double ReadHMS(char *s);
double ReadDMS(char *s);
void ReadDate(char *s, TDATE * Date);

/* Output functions */

void WriteHMS(double Time, int n);
void WriteDMS(double Angle, int n);
void WriteDate(TDATE *Date);
void cls(void);
void WaitForKey(void);

#else

/* Input functions */

int ReadInt();
double ReadReal(), ReadHMS(), ReadDMS();
void ReadDate();

/* Output functions */

void WriteHMS(), WriteDMS(), WriteDate();
void cls(), WaitForKey();

#endif

#ifdef __cplusplus
}
#endif
