/*-----------------------------------------------------------------------*/
/*  (C) Copyright Christian Monstein HB9SCT 8807 Freienbach              */
/*-----------------------------------------------------------------------*/
/*  Programmname: HELIO.CPP                                              */
/*                                                                       */
/*  Ausgabe: V2.1     Datum: 01.04.1999    Autor: Chr. Monstein          */
/*                                                                       */
/*  Zweck: Berechnet nach MEEUS die heliozentrischen Koo der Erde und LHA*/
/*                                                                       */
/*  Source: HELIO.CPP    Compiler: C++5.01                               */
/*                                                                       */
/*-----------------------------------------------------------------------*/

// Created by: Jeffrey Sax   01.11.1992   MEEUS...
// Updated by: Chr. Monstein 06.04.1997   reduziert fr MUON-Telescope
// Updated by: Chr. Monstein 01.05.1998   converted to C++
// Updated by: Chr. Monstein 01.04.1999   JD included
// Updated by: Chr. Monstein 29.11.2008   Sunrise, Sunset included
//-------------------------------------------------------------------

#include <stdio.h>
#include <math.h>

#include "helio.h"
#include "aa.h"

//-------------------------------------------------------------------
void AstroCalc(double Longitude, double Latitude, double TZ, double H0, int planet,
               int year, int month, int day, double hh,
               double *lha, double *azi, double *ele,
               double *rightascension, double *declination, double *JulianDay,
               double *s_rise, double *s_trans, double *s_set)
{

  TDATE Date;
  int Planet, Earth, RiseSetFlags;
  double ObsLon, ObsLat, Height;
  double RhoCosPhi, RhoSinPhi;
  double T, DTOfUT, DeltaT, LAST;
  double Obl, NutLon, NutObl, TimeZone;
  TSVECTOR SHelio, SEarth, SGeo;
  double RA, Decl;
  double RA1, Decl1, RA2, Decl2;
  double Az, Alt, dAlt;
  double h0, Rise, Transit, Setting;
  double Phase, PhaseAngle;
  double SemiDiameter, PolarSemiDiameter;
  double SaturnDeltaU, SaturnB, Magnitude;
  double tempJD;

  const char *PlanetName[10] = {"Sonne", "Merkur", "Venus", "Erde", "Mars",
    "Jupiter", "Saturn", "Uranus", "Neptun", "Pluto"};

  TSUNPHYSDATA SunPhysData;
  TMARSPHYSDATA MarsPhysData;
  TJUPITERPHYSDATA JupiterPhysData;
  TSATURNRINGDATA SaturnRingData;

  ObsLon = Longitude*DToR;
  ObsLat = Latitude*DToR;
  TimeZone = TZ/24.0;
  Height = H0;
  Planet = planet;
  Date.yy=year;  // Jahr
  Date.mm=month; // Monat
  Date.dd=day;   // Ganzer Tag
  Date.dd=Date.dd+hh/24.0; // plus angebrochener Tag (Stunden)

  // Wenn Planet = Pluto, brauchen wir die Position der Erde
  // fr Ekliptik und Aequinoktium von J2000, sonst des Datums.
  if (Planet == 9)
    Earth = EPJ2000;
  else
    Earth = EPDATE;

  //printf("Ephemeride fuer %s am ", PlanetName[Planet]);
  //WriteDate(&Date);
  //printf(" um ");
  //WriteHMS((Date.dd - floor(Date.dd)) * pi2, 4);
  //printf("\nLaenge = ");
  //WriteDMS(ObsLon, 4);
  //printf(", Breite = ");
  //WriteDMS(ObsLat, 4);
  //printf(", Hoehe ueber Meeresspiegel = %.0fm\n", Height);

  // vorlaeufige Berechnungen
  tempJD = CalendarToJD (&Date);
  *JulianDay = tempJD;
  T = JDToT (tempJD + TimeZone);
  DeltaT = ApproxDeltaT (T);
  DTOfUT = T + DeltaT * secToT;
  //printf("DeltaT : %4.1f Sekunden\n", DeltaT);
  InitPlanetaryTerms(0);
  if (Planet != 9)
  {
    ReadPlanetaryTerms(Planet);
    Obl = Obliquity(T);
  }
  else
    Obl = Obl2000;

  if (Planet != 3) ReadPlanetaryTerms (Earth);
  NutationConst (T, &NutLon, &NutObl);
  LAST = SiderealTime (DTOfUT) + NutLon * cos (Obl) - ObsLon;
  while (LAST>pi2) LAST=LAST-pi2;
  *lha=LAST; // Rckgabewert local sideral time

  ObserverCoord (ObsLat, Height, &RhoCosPhi, &RhoSinPhi);

  // Hauptberechnungen
  if (Planet == 0)
    SHelio.l = SHelio.b = SHelio.r = 0.0;
  else
  {
    PlanetPosHi (Planet, T, &SHelio);
    //puts("heliozentrische ekliptikale Koordinaten :");
    //printf(" l = ");
    //WriteDMS(SHelio.l + NutLon, 5);
    //printf(", b = ");
    //WriteDMS(SHelio.b, 4);
    //printf(", r = %9.6f\n", SHelio.r);
  }

  if (Planet != 3) // Erde!
  {
    PlanetPosHi (Earth, T, &SEarth);
    HelioToGeo (&SHelio, &SEarth, &SGeo);
    //puts ("geometrische geozentrische ekliptikale Koordinaten :");
    //printf (" l = ");
    //WriteDMS(SGeo.l, 5);
    //printf (", b = ");
    //WriteDMS (SGeo.b, 4);
    //printf (", Delta = %9.6f\n", SGeo.r);

    /* nur eine Lichtzeititeration */
    if (Planet != 0)
    {
      PlanetPosHi (Planet, T - SGeo.r * LightTimeConst, &SHelio);
      HelioToGeo (&SHelio, &SEarth, &SGeo);
    }

    //puts ("mittlere geozentrische ekliptikale Koordinaten :");
    //printf (" l = ");
    //WriteDMS (SGeo.l, 5);
    //printf (", b = ");
    //WriteDMS (SGeo.b, 4);
    //printf (", Delta = %9.6f\n", SGeo.r);

    EclToEqu (SGeo.l, SGeo.b, Obl, &RA, &Decl);
    //printf ("mittlere aequatoriale Koordinaten : alpha = ");
    //WriteHMS (RA, 6);
    //printf (", delta = ");
    //WriteDMS (Decl, 5);
    //putchar ('\n');

    EclToEqu (SGeo.l + NutLon, SGeo.b, Obl + NutObl, &RA, &Decl);
    Aberration (T, Obl, FK5System, &RA, &Decl);
    //printf ("scheinbare aequatoriale Koordinaten : alpha = ");
    //WriteHMS (RA, 6);
    //printf (", delta = ");
    //WriteDMS (Decl, 5);
    //putchar ('\n');

    Parallax (SolarParallax / SGeo.r, LAST, RhoCosPhi, RhoSinPhi, &RA, &Decl);
    //printf ("topozentrische Koordinaten : alpha = ");
    //WriteHMS (RA, 5);
    //printf (", delta = ");
    //WriteDMS(Decl, 4);
    //putchar('\n');
    *rightascension=RA;
    *declination=Decl;

    EquToHor (RA, Decl, LAST, ObsLat, &Az, &Alt);
    //printf ("Horizontakoordinaten : A = ");
    //WriteDMS (Az, 4);
    //printf (", h = ");
    //WriteDMS (Alt, 4);
    //putchar ('\n');
    dAlt=0; // keine Refraktion aus Prinzip
    if (Alt > -h0Planet)
    {
      dAlt = AtmRefraction (Alt);
      //printf ("Hoehe unter Einbeziehung der atmospaerischen Refraktion : ");
      //WriteDMS (Alt + dAlt, 4);
      //putchar ('\n');
    }
    *azi=Az*RToD+180.0; // Azimut
    *ele=(Alt+dAlt)*RToD; // Elevation

    //puts("\nPhysische Ephemeride : ");
    switch (Planet)
    {
      case 0:
        SunPhysEphemeris(DTOfUT, SGeo.l, Obl, NutLon, &SunPhysData);
        //printf ("  L0 = %6.2f, B0 = %5.2f, P = %5.2f\n",
        //  SunPhysData.L0 * RToD, SunPhysData.B0 * RToD, SunPhysData.P * RToD);
        break;
      case 4:
        MarsPhysEphemeris (DTOfUT, &SHelio, &SEarth, &SGeo, Obl, NutLon, NutObl, &MarsPhysData);
        //printf("  DE = %4.2f, DS = %4.2f\n",
        //  MarsPhysData.DE * RToD, MarsPhysData.DS * RToD);
        //printf("  Positionswinkel : %4.2f\n", MarsPhysData.P * RToD);
        //printf("  Laenge des Zentralmeridians : %4.2f\n",
        //  MarsPhysData.Om * RToD);
        //printf("  Lichtdefekt : %4.2f\", Positionswinkel : %4.2f\n",
        //MarsPhysData.qq, MarsPhysData.Q * RToD);
        break;
      case 5:
        JupiterPhysEphemeris(DTOfUT, &SHelio, &SEarth, &SGeo, Obl, NutLon, NutObl,&JupiterPhysData);
        //printf("  DE = %4.2f, DS = %4.2f, Positionswinkel : %4.2f\n",
        //  JupiterPhysData.DE * RToD, JupiterPhysData.DS * RToD,
        //  JupiterPhysData.P * RToD);
        //puts("  Laenge des Zentralmeridians :");
        //printf("    System I  : %4.2f, System II : %4.2f\n",
        //  JupiterPhysData.Om1 * RToD, JupiterPhysData.Om2 * RToD);
        //printf("    Phasenkorrektur : %4.2f\n", JupiterPhysData.C * RToD);
        break;
      case 6:
        SaturnRing (DTOfUT, &SHelio, &SGeo, Obl, NutLon, NutObl, &SaturnRingData);
        //printf ("  Breite der Erde  B  = %5.3f\n", SaturnRingData.B*RToD);
        //printf ("  Breite der Sonne B' = %5.3f\n", SaturnRingData.Bd*RToD);
        //printf ("  Positionswinkel : %5.3f\n", SaturnRingData.P * RToD);
        //printf ("  grosse Halbachse des Rings = %4.2f\"\n", SaturnRingData.aAxis);
        //printf ("  kleine Halbachse des Rings = %4.2f\"\n", SaturnRingData.bAxis);
        SaturnB = SaturnRingData.B;
        SaturnDeltaU = SaturnRingData.DeltaU;
        break;
    }
    if ((Planet != 0) && (Planet != 3))
    {
      Phase = CalcPhase (SHelio.r, SEarth.r, SGeo.r);
      PhaseAngle = acos (2 * Phase - 1);
      Magnitude = PlanetMagnitude(Planet, SHelio.r, SGeo.r, PhaseAngle,
      SaturnDeltaU, SaturnB);
      //printf ("  Phase : %5.3f, scheinbare Helligkeit : %4.2f\n", Phase, Magnitude);
    }
    if (Planet != 3)
    {
      SemiDiameter = PlanetSemiDiameter(Planet, SGeo.r, &PolarSemiDiameter);
      switch (Planet)
      {
        case 0 :
          //printf ("  scheinbarer Winkeldurchmesser : ");
          //WriteDMS (2*SemiDiameter * SToR, 4);
          //putchar ('\n');
          break;
        case 6 :
          CorrectSaturnSemiDiameter (SaturnB, &PolarSemiDiameter);
        case 5 :
          //printf("  scheinbarer aequatorialer Winkeldurchmesser : %4.2f\"\n", 2*SemiDiameter);
          //printf("  scheinbarer Poldurchmesser                  : %4.2f\"\n", 2*PolarSemiDiameter);
          ;
          break;
        default :
          //printf("  scheinbarer Winkeldurchmesser : %4.2f\"\n", 2*SemiDiameter);
          ;
      } // end switch
    } // end if Planet...

    // Auf- und Untergangszeiten
    // Finde T fr 0h Ortszeit
    T = (floor(T * 36525 - 0.5 - TimeZone) + 0.5 + TimeZone) / 36525;
    DTOfUT = T + secToT * DeltaT;
    if (Planet > 4)// Aenderung in RA und Dekl ist zu vernachlssigen
    {
      RA1 = RA; Decl1 = Decl;
      RA2 = RA; Decl2 = Decl;
    }
    else
    {
      if (Planet) PlanetPosHi (Planet, DTOfUT - 1/36525.0, &SHelio);
      PlanetPosHi (Earth, DTOfUT - 1/36525.0, &SEarth);
      HelioToGeo (&SHelio, &SEarth, &SGeo);
      EclToEqu (SGeo.l, SGeo.b, Obl, &RA1, &Decl1);

      if (Planet) PlanetPosHi (Planet, DTOfUT, &SHelio);
      PlanetPosHi (Earth, DTOfUT, &SEarth);
      HelioToGeo (&SHelio, &SEarth, &SGeo);
      EclToEqu (SGeo.l, SGeo.b, Obl, &RA, &Decl);

      if (Planet) PlanetPosHi (Planet, DTOfUT + 1/36525.0, &SHelio);
      PlanetPosHi (Earth, DTOfUT + 1/36525.0, &SEarth);
      HelioToGeo (&SHelio, &SEarth, &SGeo);
      EclToEqu (SGeo.l, SGeo.b, Obl, &RA2, &Decl2);
    }
    if (Planet == 0)
      h0 = h0Sun;
    else
      h0 = h0Planet;

    RiseSet (T, DeltaT,
       RA1, Decl1, RA, Decl, RA2, Decl2,
       h0, ObsLon, ObsLat, &Rise, &Transit, &Setting, &RiseSetFlags);
    switch (RiseSetFlags)
    {
      case ALWAYS_ABOVE:
        //printf ("Immer ueber dem Horizont.  Durchgang um ");
        //WriteHMS (Transit, 2);
        break;
      case ALWAYS_BELOW:
        //printf ("Immer unter dem Horizont.  Durchgang um ");
        //WriteHMS (Transit, 2);
        break;
      default:
        //printf("Aufgang um ");
        //WriteHMS (Rise, 2);
        while (Rise>pi2) Rise=Rise-pi2;
        *s_rise=Rise;
        //printf (", Durchgang um ");
        //WriteHMS (Transit, 2);
        while (Transit>pi2) Transit=Transit-pi2;
        *s_trans=Transit;
        //printf (", Untergang um ");
        //WriteHMS (Setting, 2);
        while (Setting>pi2) Setting=Setting-pi2;
        *s_set=Setting;
        ;
    }
    //putchar('\n');
  } // end if Planet!=3
  else // erde
  {
    *rightascension=0.0;
    *declination=0.0;
  }
  FreeAllPlanetaryTerms ();
}
//-----------------------------------------------------------------